/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"

bool is_sorted(int a[], int n) {
  for (int i = 0; i < n-1; i++) {
    if (a[i] > a[i+1]) {
      return false;
    }
  }
  return true;
}

// échange des valeurs de deux cases d'un tableau
void swap(int a[], int i, int j) {
  int tmp = a[i];
  a[i] = a[j];
  a[j] = tmp;
}

// mélange probabiliste
void knuth_shuffle(int a[], int n) {
  for (int i = 1; i < n; i++) {
    swap(a, i, rand() % (i+1));
  }
}

// on suppose ici que le tableau contient uniquement les valeurs 0, 1
void two_way_sort(int a[], int n) {
  int lo = 0, hi = n - 1;
  while (lo < hi) {
    if (a[lo] == 0) { lo++; }
    else if (a[hi] == 1) { hi--; }
    else { swap(a, lo++, hi--); }
  }
}

// drapeau hollandais
// on suppose ici que le tableau contient uniquement les valeurs 0, 1, 2
//    0      b       i       r     n
//   +------+-------+-------+-----+
//   |   0  |   1   | ????? |  2  |
//   +------+-------+-------+-----+
void dutch_flag(int a[], int n) {
  int b = 0, i = 0, r = n;
  while (i < r) {
    if (a[i] == 0) { swap(a, b++, i++); }
    else if (a[i] == 1) { i++; }
    else { swap(a, --r, i); }
  }
}


// quicksort, avec une partition en 3 (c'est le drapeau hollandais)
// et un mélange pour démarrer

//    l      lo      i       hi     r
//   +------+-------+-------+------+
//   |  <p  |  =p   | ????? |  >p  |
//   +------+-------+-------+------+

void quickrec0(int a[], int l, int r) { // trie uniquement a[l..r[
  if (r - l <= 1) return;
  int p = a[l], lo = l, hi = r;
  for (int i = l+1; i < hi; ) {
    if (a[i] == p) {
      i++;
    } else if (a[i] < p) {
      swap(a, i++, lo++);
    } else { // p < a[i]
      swap(a, i, --hi);
    }
  }
  quickrec0(a, l, lo);
  quickrec0(a, hi, r);
}

// encore mieux : sans faire déborder la pile

void quickrec(int a[], int l, int r) { // trie uniquement a[l..r[
  if (r - l <= 1) return;
  int p = a[l], lo = l, hi = r;
  for (int i = l+1; i < hi; ) {
    if (a[i] < p) {
      swap(a, i++, lo++);
    } else if (a[i] == p) {
      i++;
    } else { // a[i] > p
      swap(a, i, --hi);
    }
  }
  if (lo - l < r - hi) {
    quickrec(a, l, lo);
    quickrec(a, hi, r);
  } else {
    quickrec(a, hi, r);
    quickrec(a, l, lo);
  }
}

void quicksort(int a[], int n) {
  knuth_shuffle(a, n);
  quickrec(a, 0, n);
}


// quick select (a.k.a. Hoare's find)
// on donne ici une version basée sur une partition en 2

//    l         lo      hi     r
//   +-+-------+-------+------+
//   |p|  <=p  | ????? |  >p  |
//   +-+-------+-------+------+

int findrec(int k, int a[], int l, int r) {
  // cherche l'élément de rang (k-l) dans a[l..r[
  assert (l <= k && k < r);
  int p = a[l], lo = l+1, hi = r;
  while (lo < hi) {
    if (p < a[lo]) { swap(a, lo, --hi); }
    else { lo++; }
  }
  swap(a, l, lo-1);
  if      (k < lo-1)   return findrec(k, a, l, lo-1);
  else if (k >= lo)    return findrec(k, a, lo, r);
  else /* k == lo-1 */ return p;
}

int find(int k, int a[], int n) {
  assert (0 <= k && k < n);
  return findrec(k, a, 0, n);
}

// différentes variantes de bipartition
int bipartition1(int a[], int n) {
  assert (0 < n);
  int p = a[0], lo = 1, hi = n;
  while (lo < hi) {
    if (p < a[lo]) { swap(a, lo, --hi); }
    else { lo++; }
  }
  swap(a, 0, lo - 1);
  return lo - 1;
}

int bipartition2(int a[], int n) {
  assert (0 < n);
  int p = a[0], lo = 1, hi = n - 1;
  while (lo <= hi) {
    if (a[lo] <= p) { lo++; }
    else if (a[hi] >= p) { hi--; }
    else { swap(a, lo++, hi--); }
  }
  swap(a, 0, lo-1);
  return lo-1;
}

int bipartition3(int a[], int n) {
  assert (0 < n);
  int p = a[0], lo = 1, hi = n-1;
  while (lo <= hi) {
    while (lo < n && a[lo] <= p) { lo++; }
    while (hi > 0 && a[hi] >= p) { hi--; }
    if (lo < hi) { swap(a, lo++, hi--); }
  }
  swap(a, 0, lo-1);
  return lo-1;
}
